#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Author: Michael Ströder <michael@stroeder.com>

see slapd-sock(5):
https://www.openldap.org/software/man.cgi?query=slapd-sock
"""

__all__ = [
    'SlapdSockHandler',
    'NoopHandler',
]

#-----------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------

# from Python's standard lib
import sys, pwd, struct
import time
import socket, SocketServer
import logging, inspect

# internal import from slapsock
import slapdsock.message
from slapdsock.message import ENTRYResponse, RESULTResponse, InternalErrorResponse

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------


SO_PEERCRED_DICT = {
    'linux2': (17, '3i'),  # for Linux systems
}


#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------


class SlapdSockHandlerError(Exception):
    """
    Exception class to be raised within handler methods
    """

    def __init__(
            self,
            orig_exc,
            log_level=None,
            response=None,
            log_vars=None,
    ):
        Exception.__init__(self)
        self.orig_exc = orig_exc
        self.log_level = log_level
        self.log_args = None
        self.response = response
        # Store the logged vars
        log_vars = log_vars or []
        # Get the local variables dict of the caller
        frame = inspect.currentframe()
        try:
            var_dict = frame.f_back.f_locals
        finally:
            del frame
        # Store dict of only the debug vars
        self.log_vars = dict([
            (log_var, var_dict[log_var])
            for log_var in log_vars
            if log_var in var_dict
        ])

    def log(self, logger):
        """
        Write exception log message to logger
        """
        if logger.getEffectiveLevel() <= logging.DEBUG:
            # only at DEBUG level
            for var_name, var_value in self.log_vars.items():
                logger.debug('%s=%s', var_name, repr(var_value))
        if self.log_args:
            logger.log(
                self.log_level,
                *self.log_args,
                exc_info=(self.log_level >= logging.ERROR)
            )
        logger.log(
            self.log_level,
            '%s => response %s',
            repr(self.orig_exc),
            repr(self.response),
            exc_info=(self.log_level >= logging.ERROR),
        )


class SlapdSockHandler(SocketServer.BaseRequestHandler):

    """
    Base class for a handler receiving requests from
    OpenLDAP's back-sock.

    Note that when using this class as base class you have to implement
    a method for all request types hitting your handler.
    """

    cache_ttl = {}

    def __init__(self, *args, **kwargs):
        self._logged_vars = set()
        SocketServer.BaseRequestHandler.__init__(self, *args, **kwargs)

    def _get_peer_cred(self):
        """
        Currently works only on Linux
        """
        try:
            so_num, struct_fmt = SO_PEERCRED_DICT[sys.platform]
        except KeyError:
            return None, None, None
        peer_creds_struct = self.request.getsockopt(
            socket.SOL_SOCKET,
            so_num,
            struct.calcsize(struct_fmt)
        )
        pid, uid, gid = struct.unpack(struct_fmt, peer_creds_struct)
        return pid, uid, gid  # _get_peer_cred()

    def _check_access(self, uid=None):
        """
        Check whether the POSIX-UID shall be granted access
        """
        if self.peer_uid in self.server._allowed_uids:
            self._log(
                logging.DEBUG,
                'Access granted for peer UID %d',
                self.peer_uid
            )
            result = True
        elif self.peer_gid in self.server._allowed_gids:
            self._log(
                logging.DEBUG,
                'Access granted for peer GID %d',
                self.peer_gid
            )
            result = True
        elif uid != None:
            # Try to map numeric peer's uidNumber to known system user
            try:
                peer_pwd = pwd.getpwuid(self.peer_uid)
            except KeyError:
                self._log(
                    logging.WARN,
                    'Username for peer UID %d not found',
                    self.peer_uid
                )
                result = False
            else:
                # Now we know the username of the peer's UID
                self._log(
                    logging.DEBUG,
                    'peer_pwd.pw_name = %s',
                    repr(peer_pwd.pw_name)
                )
                result = peer_pwd.pw_name == uid
        else:
            result = False
        return result  # _check_access()

    def _log(self, log_level, *args, **kwargs):
        """
        Wrapper method adding log prefix
        """
        # Log variables
        if self.server.logger.getEffectiveLevel() <= logging.DEBUG:
            frame = inspect.currentframe()
            try:
                all_vars = frame.f_back.f_locals
            finally:
                del frame
            for var_name in self.server._log_vars:
                if var_name in all_vars and not var_name in self._logged_vars:
                    self._logged_vars.add(var_name)
                    self._log(logging.DEBUG, '%s=%s', var_name, repr(all_vars[var_name]))
        # Log the real stuff
        self.server.logger.log(
            log_level,
            ' '.join((self.log_prefix, args[0])),
            *args[1:],
            **kwargs
        )

    def do_monitor(self, request):
        """
        Return ENTRY response with monitoring data in response
        to MONITOR request
        """
        # Check authorization of calling process based on Unix peer credentials
        if not self._check_access():
            error_message = 'Peer %d/%d not authorized to query monitor' % (
                self.peer_uid, self.peer_gid)
            self._log(logging.ERROR, error_message)
            return RESULTResponse(
                request.msgid,
                'insufficientAccessRights',
                info=error_message
            )
        monitor_entry = self.server.monitor_entry()
        self._log(
            logging.DEBUG,
            'self.server.monitor_dn=%s',
            repr(self.server.monitor_dn),
        )
        self._log(
            logging.DEBUG,
            'monitor_entry=%s',
            repr(monitor_entry),
        )
        return ENTRYResponse(request.msgid, self.server.monitor_dn, monitor_entry)
        # end of do_monitor ()

    def handle(self):
        """
        Handle the incoming request
        """
        self.request_timestamp = time.time()
        self.server._req_count += 1
        msgid = None
        # Generate basic log prefix here
        self.log_prefix = str(id(self))
        reqtype = '-/-'

        try: # -> Exception
            self.peer_pid, self.peer_uid, self.peer_gid = self._get_peer_cred()
            self._log(
                logging.DEBUG,
                '----- incoming request via %s from pid=%s uid=%s gid=%s -----',
                repr(self.request.getsockname()),
                self.peer_pid, self.peer_uid, self.peer_gid,
            )
            request_data = self.request.recv(500000)
            if __debug__:
                # Security advice:
                # Request data can contain clear-text passwords!
                self._log(logging.DEBUG, 'request_data=%s', repr(request_data))
            self.server._bytes_received += len(request_data)
            req_lines = request_data.split('\n')
            # Extract request type
            reqtype = req_lines[0]
            self._log(logging.DEBUG, 'reqtype=%s', repr(reqtype))
            # Update request counter for request type
            self.server._req_counters[reqtype.lower()] += 1
            # Get the request message class
            request_class = getattr(slapdsock.message, '%sRequest' % reqtype)
            # Extract the request message
            sock_req = request_class(req_lines)
            if __debug__:
                # Security advice:
                # Request data can contain sensitive data
                # (e.g. BIND with password) => never run in debug mode!
                self._log(
                    logging.DEBUG,
                    'sock_req=%s // %s',
                    repr(sock_req),
                    sock_req.__dict__
                )
            # Generate the request specific log prefix here
            self.log_prefix = sock_req.log_prefix(self.log_prefix)
            msgid = sock_req.msgid
            cache_key = sock_req.cache_key()

            try: # -> SlapdSockHandlerError

                try: # Try cache
                    response = self.server.req_cache[reqtype][cache_key]
                except KeyError:
                    self._log(
                        logging.DEBUG,
                        'Request not cached: cache_key=%s',
                        repr(cache_key),
                    )
                    # Get the handler method in own class
                    handle_method = getattr(self, 'do_%s' % reqtype.lower())
                    # Let the handler method generate a response message
                    response = handle_method(sock_req)
                    if cache_key and reqtype in self.server.req_cache:
                        # Store response in cache
                        self.server.req_cache[reqtype][cache_key] = response
                        self._log(
                            logging.DEBUG,
                            'Response stored in cache: cache_key=%s',
                            repr(cache_key),
                        )

                else:
                    self._log(
                        logging.DEBUG,
                        'Response retrieved from cache: cache_key=%s',
                        repr(cache_key),
                    )

            except SlapdSockHandlerError, handler_exc:
                handler_exc.log(self.server.logger)
                response = handler_exc.response or InternalErrorResponse(msgid)

        except Exception:
            self._log(
                logging.ERROR,
                'Unhandled exception during processing request:',
                exc_info=True
            )
            response = InternalErrorResponse(msgid)
        try:
            # Serialize the response instance
            response_str = str(response)
            self._log(logging.DEBUG, 'response_str=%s', repr(response_str))
            if response_str:
                self.request.sendall(response_str)
        except Exception:
            self._log(
                logging.ERROR,
                'Unhandled exception while sending response:',
                exc_info=True
            )
        else:
            response_delay = time.time()-self.request_timestamp
            self.server.update_monitor_data(len(response_str), response_delay)
            self._log(logging.DEBUG, 'response_delay=%0.3f', response_delay)
        return  # handle()


class NoopHandler(SlapdSockHandler):

    """
    This handler simply returns CONTINUE+LF for every sockops request
    and empty string for sockresps and unbind requests.

    This is handy to be used as safe base class for own custom handler
    to make sure each back-sock request is always answered in
    case of misconfigured "overlay sock" section.
    """

    def do_add(self, request):
        """
        ADD
        """
        _ = (self, request) # pylint dummy
        return 'CONTINUE\n'

    def do_bind(self, request):
        """
        BIND
        """
        _ = (self, request) # pylint dummy
        return 'CONTINUE\n'

    def do_compare(self, request):
        """
        COMPARE
        """
        _ = (self, request) # pylint dummy
        return 'CONTINUE\n'

    def do_delete(self, request):
        """
        DELETE
        """
        _ = (self, request) # pylint dummy
        return 'CONTINUE\n'

    def do_modify(self, request):
        """
        MODIFY
        """
        _ = (self, request) # pylint dummy
        return 'CONTINUE\n'

    def do_modrdn(self, request):
        """
        MODRDN
        """
        _ = (self, request) # pylint dummy
        return 'CONTINUE\n'

    def do_search(self, request):
        """
        SEARCH
        """
        _ = (self, request) # pylint dummy
        return 'CONTINUE\n'

    def do_unbind(self, request):
        """
        UNBIND
        """
        _ = (self, request) # pylint dummy
        return ''

    def do_result(self, request):
        """
        RESULT
        """
        _ = (self, request) # pylint dummy
        return ''

    def do_entry(self, request):
        """
        ENTRY
        """
        _ = (self, request) # pylint dummy
        return ''
