#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Author: Michael Ströder <michael@stroeder.com>
"""

__all__ = [
    'RESULT_CODE_NAME',
    'RESULT_CODE',
    'MyLDAPUrl',
    'MyLDAPObject',
]

#-----------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------

# from standard lib
import time, datetime

# from package python-ldap
import ldap
import ldapurl
import ldap.resiter
import ldap.sasl
from ldap.ldapobject import ReconnectLDAPObject


#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------

# Number of times connecting to LDAP is retried before sending a failed
# response for a query
LDAP_MAXRETRYCOUNT = 10
# Time to wait before retrying to connect within one query
LDAP_RETRYDELAY = 0.1

# Timeout in seconds when connecting to local and remote LDAP servers
# used for ldap.OPT_NETWORK_TIMEOUT and ldap.OPT_TIMEOUT
LDAP_TIMEOUT = 3.0

RESULT_CODE_NAME = {
    0: 'success',
    1: 'operationsError',
    2: 'protocolError',
    3: 'timeLimitExceeded',
    4: 'sizeLimitExceeded',
    5: 'compareFalse',
    6: 'compareTrue',
    7: 'authMethodNotSupported',
    8: 'strongerAuthRequired',
    #     -- 9 reserved --
    10: 'referral',
    11: 'adminLimitExceeded',
    12: 'unavailableCriticalExtension',
    13: 'confidentialityRequired',
    14: 'saslBindInProgress',
    16: 'noSuchAttribute',
    17: 'undefinedAttributeType',
    18: 'inappropriateMatching',
    19: 'constraintViolation',
    20: 'attributeOrValueExists',
    21: 'invalidAttributeSyntax',
    32: 'noSuchObject',
    33: 'aliasProblem',
    34: 'invalidDNSyntax',
    #     -- 35 reserved for undefined isLeaf --
    36: 'aliasDereferencingProblem',
    #     -- 37-47 unused --
    48: 'inappropriateAuthentication',
    49: 'invalidCredentials',
    50: 'insufficientAccessRights',
    51: 'busy',
    52: 'unavailable',
    53: 'unwillingToPerform',
    54: 'loopDetect',
    #     -- 55-63 unused --
    64: 'namingViolation',
    65: 'objectClassViolation',
    66: 'notAllowedOnNonLeaf',
    67: 'notAllowedOnRDN',
    68: 'entryAlreadyExists',
    69: 'objectClassModsProhibited',
    #     -- 70 reserved for CLDAP --
    71: 'affectsMultipleDSAs',
    #     -- 72-79 unused --
    80: 'other',
}

RESULT_CODE = dict([
    (v, k)
    for k, v in RESULT_CODE_NAME.items()
])


RESULT_CODE.update({
    ldap.SUCCESS: 0x00,
    ldap.OPERATIONS_ERROR: 0x01,
    ldap.PROTOCOL_ERROR: 0x02,
    ldap.TIMELIMIT_EXCEEDED: 0x03,
    ldap.SIZELIMIT_EXCEEDED: 0x04,
    ldap.COMPARE_FALSE: 0x05,
    ldap.COMPARE_TRUE: 0x06,
    ldap.AUTH_METHOD_NOT_SUPPORTED: 0x07,
    ldap.STRONG_AUTH_REQUIRED: 0x08,
    ldap.PARTIAL_RESULTS: 0x09,
    ldap.ADMINLIMIT_EXCEEDED: 0x0b,
    ldap.CONFIDENTIALITY_REQUIRED: 0x0d,
    ldap.NO_SUCH_ATTRIBUTE: 0x10,
    ldap.UNDEFINED_TYPE: 0x11,
    ldap.INAPPROPRIATE_MATCHING: 0x12,
    ldap.CONSTRAINT_VIOLATION: 0x13,
    ldap.TYPE_OR_VALUE_EXISTS: 0x14,
    ldap.INVALID_SYNTAX: 0x15,
    ldap.NO_SUCH_OBJECT: 0x20,
    ldap.ALIAS_PROBLEM: 0x21,
    ldap.INVALID_DN_SYNTAX: 0x22,
    ldap.IS_LEAF: 0x23,
    ldap.ALIAS_DEREF_PROBLEM: 0x24,
    ldap.X_PROXY_AUTHZ_FAILURE: 0x2F,
    ldap.INAPPROPRIATE_AUTH: 0x30,
    ldap.INVALID_CREDENTIALS: 0x31,
    ldap.INSUFFICIENT_ACCESS: 0x32,
    ldap.BUSY: 0x33,
    ldap.UNAVAILABLE: 0x34,
    ldap.UNWILLING_TO_PERFORM: 0x35,
    ldap.LOOP_DETECT: 0x36,
    ldap.NAMING_VIOLATION: 0x40,
    ldap.OBJECT_CLASS_VIOLATION: 0x41,
    ldap.NOT_ALLOWED_ON_NONLEAF: 0x42,
    ldap.NOT_ALLOWED_ON_RDN: 0x43,
    ldap.ALREADY_EXISTS: 0x44,
    ldap.NO_OBJECT_CLASS_MODS: 0x45,
    ldap.RESULTS_TOO_LARGE: 0x46,
    ldap.AFFECTS_MULTIPLE_DSAS: 0x47,
    ldap.VLV_ERROR: 0x4C,
    ldap.OTHER: 0x50,
})


#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------

LDAP_DATETIME_FORMAT = r'%Y%m%d%H%M%SZ'


def ldap_float_str(fln):
    """
    Return fln as string formatted for NumString
    """
    return '%0.5f' % fln


def ldap_datetime(time_str):
    """
    Parse GeneralizedTime (UTC) string to datetime.datetime instance
    """
    return datetime.datetime.strptime(time_str, LDAP_DATETIME_FORMAT)


def ldap_datetime_str(time_dt):
    """
    Return datetime as string formatted as GeneralizedTime (UTC)
    """
    return datetime.datetime.strftime(time_dt, LDAP_DATETIME_FORMAT)


def is_expired(change_timestamp_str, max_age_str, now_dt=None):
    """
    Returns True if change_timestamp_str + max_age_str > now_dt
    """
    max_age_int = int(max_age_str)
    if max_age_int <= 0 or not change_timestamp_str:
        # nothing to check
        return False
    now_dt = now_dt or datetime.datetime.utcnow()
    max_age_td = datetime.timedelta(seconds=int(max_age_str))
    change_timestamp_dt = ldap_datetime(change_timestamp_str)
    expire_dt = change_timestamp_dt + max_age_td
    return expire_dt < now_dt


class MyLDAPUrl(ldapurl.LDAPUrl):
    """
    Extended LDAPUrl
    """
    attr2extype = {
        'who': 'bindname',
        'cred': 'X-BINDPW',
        'trace_level': 'trace',
        'pwd_filename': 'x-pwdfilename',
        'sasl_mech': 'x-saslmech',
    }


class MyLDAPObject(ReconnectLDAPObject, ldap.resiter.ResultProcessor):
    """
    Extended LDAPObject
    """

    # Number of seconds search requests are cached
    cache_time = 0.0

    def __init__(
            self, ldap_url,
            # Trace arguments for debugging
            trace_level=0, trace_file=None, trace_stack_limit=5,
            # Retry arguments, see ReconnectLDAPObject
            retry_max=LDAP_MAXRETRYCOUNT, retry_delay=LDAP_RETRYDELAY,
            timeout=LDAP_TIMEOUT,
            # Simple bind arguments
            who=None,
            cred=None,
            # TLS arguments
            cacert_filename=None,
            client_cert_filename=None,
            client_key_filename=None,
            cache_time=30.0,
        ):
        """
        Opens connection, sets some LDAP options and binds according
        to what's provided in `uri'.

        Extensions to parameters passed to ReconnectLDAPObject():
        `ldap_url'
          Can contain the full LDAP URL with authc information and base-DN
        `who'
          Bind-DN to be used (overrules ldap_url)
        `cred'
          Bind password to be used (overrules ldap_url)
        """
        # Parse LDAP URL
        self.ldap_url_obj = MyLDAPUrl(ldap_url)
        who = who or self.ldap_url_obj.who
        cred = cred or self.ldap_url_obj.cred
        self.cache_time = cache_time
        # Initialize the ldap.ldapobject.ReconnectLDAPObject instance
        ReconnectLDAPObject.__init__(
            self,
            self.ldap_url_obj.initializeUrl(),
            trace_level, trace_file, trace_stack_limit,
            retry_max, retry_delay,
        )
        # Reset cache
        self.flush_cache()
        # Set TLS options
        self.set_tls_options(
            cacert_filename=cacert_filename,
            client_cert_filename=client_cert_filename,
            client_key_filename=client_key_filename,
        )
        # Switch off automatic referral chasing
        self.set_option(ldap.OPT_REFERRALS, 0)
        # Switch off automatic alias dereferencing
        self.set_option(ldap.OPT_DEREF, ldap.DEREF_NEVER)
        # Set timeout values
        self.set_option(ldap.OPT_NETWORK_TIMEOUT, timeout)
        self.set_option(ldap.OPT_TIMEOUT, timeout)
        # Send StartTLS extended operation if clear-text connection was
        # initialized
        if self.ldap_url_obj.urlscheme == 'ldap':
            self.start_tls_s()
        # Bind with authc information found in ldap_url
        if self.ldap_url_obj.sasl_mech and self.ldap_url_obj.sasl_mech.upper() == 'EXTERNAL':
            self.sasl_external_bind_s()
        elif who and cred:
            self.simple_bind_s(who, cred)
        return  # __init__()

    def set_tls_options(
            self,
            cacert_filename=None,
            client_cert_filename=None,
            client_key_filename=None,
        ):
        """
        Sets various TLS options for connection
        """
        # TLS options
        # Force server cert validation
        ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_DEMAND)
        # Set path names of TLS related files
        for tls_option, tls_pathname in (
                (ldap.OPT_X_TLS_CACERTFILE, cacert_filename,),
                (ldap.OPT_X_TLS_CERTFILE, client_cert_filename),
                (ldap.OPT_X_TLS_KEYFILE, client_key_filename),
            ):
            if not tls_pathname:
                # get global default
                tls_pathname = ldap.get_option(tls_option)
            if tls_pathname:
                # Check whether CA cert file can be read
                open(tls_pathname, 'rb')
                self._l.set_option(tls_option, tls_pathname)
        # this has to be the last option set to let libldap reinitialize
        # the TLS context
        self._l.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
        return  # set_tls_options()

    def get_cache_hit_ratio(self):
        """
        Returns percentage of cache hit ratio
        """
        try:
            return float(self._cache_hits)/float(self._cache_hits+self._cache_misses)*100
        except ZeroDivisionError:
            return None

    def flush_cache(self):
        """
        Reset the search cache
        """
        self._cache_misses = 0
        self._cache_hits = 0
        self._cache = {}

    def uncache_entry(self, dn):
        """
        Remove all occurences of entry specified with dn
        """
        dn = dn.lower()
        for cache_key_args, cached_results in self._cache.items():
            if cache_key_args[0].lower() == dn:
                del self._cache[cache_key_args]
            else:
                for cached_dn, _ in cached_results[1]:
                    if not cached_dn is None and cached_dn.lower() == dn:
                        del self._cache[cache_key_args]
                        break # stop looping through results
        return # end of uncache_entry()

    def read_s(
            self,
            dn,
            filterstr=None,
            attrlist=None,
            serverctrls=None,
            clientctrls=None,
            timeout=-1,
            nocache=0,
            cache_time=None,
        ):
        ldap_result = self.search_ext_s(
            dn,
            ldap.SCOPE_BASE,
            filterstr or '(objectClass=*)',
            attrlist=attrlist,
            serverctrls=serverctrls,
            clientctrls=clientctrls,
            timeout=timeout,
            nocache=nocache,
            cache_time=cache_time,
        )
        if ldap_result:
            return ldap_result[0][1]
        else:
            return None

    def search_ext_s(
            self,
            base,
            scope,
            filterstr='(objectClass=*)',
            attrlist=None,
            attrsonly=0,
            serverctrls=None,
            clientctrls=None,
            timeout=-1,
            sizelimit=0,
            nocache=0,
            cache_time=None,
        ):
        current_time = time.time()
        cache_key_args = (
            base,
            scope,
            filterstr,
            tuple(attrlist or []),
            attrsonly,
            tuple(serverctrls or []),
            tuple(clientctrls or []),
            timeout,
            sizelimit
        )
        try:
            result_cache_time, result = self._cache[cache_key_args]
        except KeyError:
            result = None
        else:
            if nocache or current_time > result_cache_time:
                del self._cache[cache_key_args]
                result = None
            else:
                self._cache_hits += 1
        if result is None:
            self._cache_misses += 1
            result = ReconnectLDAPObject.search_ext_s(
                self,
                base,
                scope,
                filterstr,
                attrlist,
                attrsonly,
                serverctrls,
                clientctrls,
                timeout,
                sizelimit
            )
            cache_time = cache_time or self.cache_time
            if cache_time > 0:
                self._cache[cache_key_args] = (
                    current_time+cache_time,
                    result
                )
        return result # search_ext_s()
