#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Helper stuff for using logging module

Author: Michael Ströder <michael@stroeder.com>
"""

__all__ = [
    'SYS_LOG_FORMAT',
    'CONSOLE_LOG_FORMAT',
    'LoggerFileobj',
    'combined_logger',
]

#-----------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------

# from Python's standard lib
import logging
from logging.handlers import SysLogHandler

#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------


SYS_LOG_FORMAT = '%(levelname)s %(message)s'

CONSOLE_LOG_FORMAT = '%(asctime)s %(levelname)s %(message)s'


#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------


class LoggerFileobj(object):

    """
    file-like helper class which sends log lines to a logging instance
    """

    def __init__(self, logger, logging_level):
        self._logger = logger
        self._logging_level = logging_level

    def write(self, msg):
        """
        Write :msg:str to logger.
        """
        self._logger.log(self._logging_level, msg[:-1])

    def flush(self):
        """
        dummy method
        """
        return


def combined_logger(
        log_name,
        log_level=logging.INFO,
        sys_log_format=None,
        console_log_format=None,
    ):
    """
    Returns a combined SysLogHandler/StreamHandler logging instance
    with formatters
    """

    # for writing to syslog
    new_logger = logging.getLogger(log_name)

    if sys_log_format:
        my_syslog_formatter = logging.Formatter(
            fmt=' '.join((log_name, sys_log_format)))
        my_syslog_handler = logging.handlers.SysLogHandler(
            address='/dev/log',
            facility=SysLogHandler.LOG_DAEMON,
        )
        my_syslog_handler.setFormatter(my_syslog_formatter)
        new_logger.addHandler(my_syslog_handler)

    if console_log_format:
        my_stream_formatter = logging.Formatter(fmt=console_log_format)
        my_stream_handler = logging.StreamHandler()
        my_stream_handler.setFormatter(my_stream_formatter)
        new_logger.addHandler(my_stream_handler)

    new_logger.setLevel(log_level)

    return new_logger  # end of combined_logger()
