#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Processing messages received/returned from/to OpenLDAP's back-sock

Author: Michael Ströder <michael@stroeder.com>
"""

__all__ = [
    'ListFile',
    'SockRequest',
    'ADDRequest',
    'BINDRequest',
    'COMPARERequest',
    'DELETERequest',
    'MODIFYRequest',
    'MODRDNRequest',
    'RESULTRequest',
    'ENTRYRequest',
    'SEARCHRequest',
    'UNBINDRequest',
    'SockResponse',
    'RESULTResponse',
    'ENTRYResponse',
]

#-----------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------

# from Python's standard lib
import hashlib
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from types import IntType

# from python-ldap
import ldif
from ldif import LDIFWriter

from slapdsock.ldaphelper import RESULT_CODE

#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------


def ldap_str(s):
    """
    Decode the string s to Unicode object
    """
    return s.decode('utf-8')


def ldap_attrs(s):
    """
    Return attribute list (for search requests)
    """
    if s == 'ALL':
        return u'*'
    else:
        return u' '.split(s.decode('utf-8'))


class ListFile(object):

    """
    File-like object with readline method returning lines from a list
    of strings
    """

    def __init__(self, lines, line_counter=0):
        self._lines = lines
        self._line_counter = line_counter

    def readline(self):
        """
        Read a single line without trailing space
        """
        if self._line_counter >= len(self._lines):
            line = ''
        else:
            line = self._lines[self._line_counter]
            self._line_counter += 1
        return line


class SockRequest(object):

    """
    Base class for request messages sent by OpenLDAP's back-sock
    """

    req_attrs = {
        'msgid': int,
        'binddn': ldap_str,
        'peername': ldap_str,
        'ssf': int,
        'connid': int,
        'suffix': ldap_str,
    }
    cache_key_attrs = tuple()

    def __init__(self, req_lines):
        self._req_lines = req_lines
        self.reqtype = req_lines[0]
        self._linecount = 1
        self._get_attrs()

    def _get_attrs(self):
        """
        Read the class attributes from the message line list
        """
        while self._req_lines[self._linecount]:
            try:
                k, v = self._req_lines[self._linecount].split(': ', 1)
            except ValueError:
                break
            if not k in self.req_attrs:
                break
            setattr(self, k, self.req_attrs[k](v))
            self._linecount += 1
        return

    def __str__(self):
        return '\n'.join(self._req_lines)

    def cache_key(self):
        """
        Generated a hash-able cache key from the request data.

        Do not put secure data into the cache key
        when overriding this method!
        """
        if not self.cache_key_attrs:
            return None
        cache_attrs_list = []
        for a in self.cache_key_attrs:
            try:
                cache_attrs_list.append((a, getattr(self, a)))
            except AttributeError:
                pass
        return tuple(cache_attrs_list)

    def log_prefix(self, prefix):
        """
        Return a logging prefix string
        """
        result = [prefix]
        for attr_name in ('connid', 'msgid'):
            try:
                v = getattr(self, attr_name)
            except AttributeError:
                pass
            else:
                result.append('%s=%s' % (attr_name, repr(v)))
        return ' '.join(result)


class MONITORRequest(SockRequest):

    """
    For manually injected MONITOR requests
    """

    def __init__(self, req_lines):
        SockRequest.__init__(self, req_lines)
        self.msgid = 0


class ADDRequest(SockRequest):

    """
    ADD
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    <entry in LDIF format>
    <blank line>
    """

    def _parse_ldif(self, linecount, max_entries=1):
        """
        Parse the subsequent request lines (starting from :linecount: as
        LDIF records and return all in a single list.
        """
        ldif_file = ListFile(self._req_lines, linecount)
        lrl = ldif.LDIFRecordList(ldif_file, max_entries=max_entries)
        lrl.parse_entry_records()
        return lrl.all_records

    def _get_attrs(self):
        SockRequest._get_attrs(self)
        self.dn, self.entry = self._parse_ldif(
            self._linecount, max_entries=1)[0]


class BINDRequest(SockRequest):

    """
    BIND
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    dn: <DN>
    method: <method number>
    credlen: <length of <credentials>>
    cred: <credentials>
    <blank line>
    """
    req_attrs = {
        'msgid': int,
        'binddn': ldap_str,
        'peername': ldap_str,
        'ssf': int,
        'connid': int,
        'suffix': ldap_str,
        'dn': ldap_str,
        'method': int,
        'credlen': int,
        'cred': str,
    }
    cache_key_attrs = (
        'binddn',
        'ssf',
        'suffix',
        'dn',
        'method',
    )

    def _get_attrs(self):
        SockRequest._get_attrs(self)
        cred_str_list = [self.cred]
        while self._req_lines[self._linecount]:
            cred_str_list.append(self._req_lines[self._linecount])
            self._linecount += 1
        self.cred = '\n'.join(cred_str_list)
        if len(self.cred) != self.credlen:
            raise ValueError('credlen: %d does not fit byte count of cred: %s' % (
                self.credlen, repr(self.cred),
            ))

    def cache_key(self):
        return SockRequest.cache_key(self) + \
               (('credhash', hashlib.new('sha512', self.cred).digest()), )


class COMPARERequest(SockRequest):

    """
    COMPARE
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    dn: <DN>
    <attribute>: <value>
    <blank line>
    """
    req_attrs = {
        'msgid': int,
        'binddn': ldap_str,
        'peername': ldap_str,
        'ssf': int,
        'connid': int,
        'suffix': ldap_str,
        'dn': ldap_str,
    }
    cache_key_attrs = (
        'binddn',
        'ssf',
        'suffix',
        'dn',
    )


class DELETERequest(SockRequest):

    """
    DELETE
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    dn: <DN>
    <blank line>
    """
    req_attrs = {
        'msgid': int,
        'binddn': ldap_str,
        'peername': ldap_str,
        'ssf': int,
        'connid': int,
        'suffix': ldap_str,
        'dn': ldap_str,
    }


class MODIFYRequest(ADDRequest):

    """
    MODIFY
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    dn: <DN>
    <repeat {
        <"add"/"delete"/"replace">: <attribute>
        <repeat { <attribute>: <value> }>
        -
    }>
    <blank line>
    """

    def _parse_ldif(self, linecount, max_entries=1):
        """
        Parse the subsequent request lines (starting from :linecount: as
        LDIF records and return all in a single list.
        """
        self._req_lines.insert(linecount+1, 'changetype: modify')
        ldif_file = ListFile(self._req_lines, linecount)
        lrl = ldif.LDIFRecordList(ldif_file, max_entries=max_entries)
        lrl.parse_change_records()
        return lrl.all_modify_changes

    def _get_attrs(self):
        SockRequest._get_attrs(self)
        self.dn, self.modops, _ = self._parse_ldif(
            self._linecount, max_entries=1)[0]


class MODRDNRequest(SockRequest):

    """
    MODRDN
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    dn: <DN>
    newrdn: <new RDN>
    deleteoldrdn: <0 or 1>
    <if new superior is specified: "newSuperior: <DN>">
    <blank line>
    """
    req_attrs = {
        'msgid': int,
        'binddn': ldap_str,
        'peername': ldap_str,
        'ssf': int,
        'connid': int,
        'suffix': ldap_str,
        'dn': ldap_str,
        'newrdn': ldap_str,
        'deleteoldrdn': int,
        'newSuperior': ldap_str,
    }


class RESULTRequest(SockRequest):

    """
    RESULT
    msgid: <message id>
    code: <integer>
    matched: <matched DN>
    info: <text>
    <blank line>
    """
    req_attrs = {
        'msgid': int,
        'code': int,
        'matched': ldap_str,
        'info': ldap_str,
    }


class ENTRYRequest(SockRequest):

    """
    ENTRY
    msgid: <message id>
    <entry in LDIF format>
    <blank line>
    """
    req_attrs = {
        'msgid': int,
    }


class SEARCHRequest(SockRequest):

    """
    SEARCH
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    base: <base DN>
    scope: <0-2, see ldap.h>
    deref: <0-3, see ldap.h>
    sizelimit: <size limit>
    timelimit: <time limit>
    filter: <filter>
    attrsonly: <0 or 1>
    attrs: <"all" or space-separated attribute list>
    <blank line>
    """
    req_attrs = {
        'msgid': int,
        'binddn': ldap_str,
        'peername': ldap_str,
        'ssf': int,
        'connid': int,
        'suffix': ldap_str,
        'base': ldap_str,
        'scope': int,
        'deref': int,
        'sizelimit': int,
        'timelimit': int,
        'filter': ldap_str,
        'attrs': ldap_attrs,
    }
    cache_key_attrs = (
        'binddn',
        'ssf',
        'suffix',
        'base',
        'scope',
        'deref',
        'sizelimit',
        'timelimit',
        'filter',
        'attrs',
    )


class UNBINDRequest(SockRequest):

    """
    UNBIND
    msgid: <message id>
    <repeat { "suffix:" <database suffix DN> }>
    <blank line>
    """


class SockResponse(object):

    """
    Base class for response messages returned to OpenLDAP's back-sock
    """

    line_sep = '\n'

    def __init__(self, resp_type, resp_lines=None):
        self._resp_type = resp_type
        self._resp_lines = resp_lines or []

    def __str__(self):
        lines = [self._resp_type]
        lines.extend([
            ': '.join((k, v))
            for k, v in self._resp_lines
        ])
        return self.line_sep.join(lines)


class RESULTResponse(SockResponse):

    """
    RESULT
    msgid: <message id>
    code: <integer>
    matched: <matched DN>
    info: <text>
    <blank line>
    """

    def __init__(self, msgid, code=0, matched=None, info=None):
        msgid = msgid or 0
        if isinstance(code, basestring):
            code = RESULT_CODE[code]
        assert type(msgid) == IntType, ValueError(
            'Argument msgid must be integer but was: %s' % repr(msgid)
        )
        assert type(code) == IntType, ValueError(
            'Argument code must be integer but was: %s' % repr(code)
        )
        resp_lines = [
            #('msgid', str(msgid)),
            ('code', str(code)),
        ]
        if matched != None:
            resp_lines.append(('matched', matched.encode('utf-8')))
        if info != None:
            resp_lines.append(('info', info.encode('utf-8')))
        SockResponse.__init__(self, 'RESULT', resp_lines)


class ENTRYResponse(SockResponse):

    """
    ENTRY
    msgid: <message id>
    <entry in LDIF format>
    <blank line>
    """

    def __init__(self, msgid, dn, entry):
        resp_lines = [
        ]
        self._dn = dn
        self._entry = entry
        SockResponse.__init__(self, 'ENTRY', resp_lines)

    def __str__(self):
        str_fileobj = StringIO()
        ldif_writer = LDIFWriter(str_fileobj)
        ldif_writer.unparse(self._dn, self._entry)
        ldif_str = str_fileobj.getvalue()
        str_fileobj.close()
        return '\n'.join((
            SockResponse.__str__(self),
            ldif_str
        ))


class InvalidCredentialsResponse(RESULTResponse):
    """
    Convenience wrapper class for returning invalidCredentials
    """

    def __init__(self, msgid, info=None):
        RESULTResponse.__init__(
            self,
            msgid,
            code='invalidCredentials',
            matched=None,
            info=info
        )


class InternalErrorResponse(RESULTResponse):
    """
    Convenience wrapper class for returning unwillingToPerform and 
    diagnostic message 'internal error'
    """

    def __init__(self, msgid, info=None):
        RESULTResponse.__init__(
            self,
            msgid,
            code='unwillingToPerform',
            info=info or 'internal error'
        )
