#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
The Unix domain listener

Author: Michael Ströder <michael@stroeder.com>
"""

#-----------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------

# from Python's standard lib
import grp, os, pwd, socket, SocketServer, datetime, logging, inspect, threading

# python-ldap
import ldap

from slapdsock.ldaphelper import MyLDAPObject
from slapdsock.ldaphelper import ldap_float_str, ldap_datetime_str
from slapdsock.cache import CacheDict
from slapdsock.loghelper import LoggerFileobj

#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------


class SlapdSockServer(SocketServer.UnixStreamServer):

    """
    Base class for a Unix domain socket server implementing
    an external back-sock listener
    """

    cache_ttl = -1.0
    allow_reuse_address = True
    monitor_dn = 'cn=sock-monitor'
    ldapi_uri = 'ldapi://'
    ldapi_authz_id = ''
    ldap_retry_max = 4
    ldap_retry_delay = 1.0
    ldap_cache_ttl = 0.0
    ldap_trace_level = 0

    def __init__(
            self,
            server_address,
            handler_class,
            logger,
            average_count,
            socket_timeout,
            socket_permissions,
            allowed_uids,
            allowed_gids,
            bind_and_activate=True,
            monitor_dn=None,
            log_vars=None,
        ):
        # Initialize cache per request type
        self.req_cache = {}
        for reqtype, cache_ttl in handler_class.cache_ttl.items():
            if cache_ttl > 0:
                self.req_cache[reqtype] = CacheDict(cache_ttl=cache_ttl)
        self.max_req_cache_time = 0.0
        self.logger = logger
        self._socket_timeout = socket_timeout
        self._socket_permissions = socket_permissions
        SocketServer.UnixStreamServer.__init__(
            self, server_address, handler_class, bind_and_activate
        )
        self.logger.info(
            'Initializing %s instance listening on %s',
            self.__class__.__name__,
            repr(server_address)
        )
        self._average_count = average_count
        self._start_time = datetime.datetime.utcnow()
        # Global request counter
        self._req_count = 0
        # Request counter dict per request type
        self._req_counters = dict([
            (m[3:], 0)
            for m, _ in inspect.getmembers(
                handler_class,
                predicate=inspect.ismethod
            )
            if m.startswith('do_')
        ])
        self._bytes_sent = 0
        self._bytes_received = 0
        self._avg_response_time = 0.0
        self._max_response_time = 0.0
        self._allowed_uids = self._map_names(
            pwd.getpwnam,
            pwd.getpwuid,
            allowed_uids
        )
        self._allowed_gids = self._map_names(
            grp.getgrnam,
            grp.getgrgid,
            allowed_gids
        )
        self._monitor_dn = monitor_dn or 'cn=%s' % (self.__class__.__name__)
        self._log_vars = sorted(log_vars or [])
        self.logger.debug('%s.log_vars=%s', self.__class__.__name__, log_vars)
        # Create a fileobj-like logging wrapper instance
        self._logger_fileobj = LoggerFileobj(self.logger, logging.DEBUG)
        # For lazy LDAPI connecting in self.get_ldapi_conn()
        self._ldapi_conn = None
        self._ldapi_conn_lock = ldap.LDAPLock(
            desc='get_ldapi_conn() in %s' % (repr(self.__class__))
        )
        # end of SlapdSockServer.__init__()

    def _map_names(self, map_name_func, map_id_func, nameorid_list):
        """
        Map user or group names to their POSIX id
        """
        id_set = set()
        for i in nameorid_list:
            if isinstance(i, int):
                try:
                    map_id_func(i)
                except KeyError:
                    self.logger.warn('Name for allowed ID %d not found', i)
                else:
                    id_set.add(i)
            elif isinstance(i, basestring):
                try:
                    mapped_id = map_name_func(i)[2]
                except KeyError:
                    self.logger.warn('ID for allowed name %s not found', repr(i))
                else:
                    id_set.add(mapped_id)
        self.logger.debug(
            '%s ID set: %s',
            map_name_func.__name__,
            ','.join([str(posix_id) for posix_id in id_set]),
        )
        return id_set

    def update_monitor_data(self, r_len, r_delay):
        """
        Update some monitoring data
        """
        self._bytes_sent += r_len
        self._avg_response_time = (
            (self._average_count - 1) * self._avg_response_time + r_delay
        ) / self._average_count
        self._max_response_time = max(
            self._max_response_time,
            r_delay
        )

    def monitor_entry(self):
        """
        Returns entry dictionary with monitoring data.

        Override this method to extend the monitor entry with additional
        attributes.
        """
        monitor_entry = {
            'sockPythonDebug': [str(__debug__)],
            'sockLogLevel': [str(self.logger.getEffectiveLevel())],
            'sockThreadCount': [str(threading.activeCount())],
            'sockStartTime': [ldap_datetime_str(self._start_time)],
            'sockCurrentTime': [ldap_datetime_str(datetime.datetime.utcnow())],
            # Request counters
            'sockRequestAll': [str(self._req_count)],
            # Byte counters
            'sockBytesReceived': [str(self._bytes_received)],
            'sockBytesSent': [str(self._bytes_sent)],
            # Response times
            'sockAvgResponseTime': [ldap_float_str(self._avg_response_time)],
            'sockMaxResponseTime': [ldap_float_str(self._max_response_time)],
            'sockLDAPIConnection': [repr(self._ldapi_conn)],
        }
        if self._ldapi_conn:
            monitor_entry['sockLDAPIAuthzID'] = [self._ldapi_conn.whoami_s()]
        else:
            monitor_entry['sockLDAPIAuthzID'] = ['None']
        # Counters per request type
        for req_type, req_counter in sorted(self._req_counters.items()):
            req_type_str = req_type[0].upper(), req_type[1:].lower()
            monitor_entry['sockRequest%s%sCount' % ((req_type_str))] = [str(req_counter)]
        # Caches per request type
        for req_type, req_cache in sorted(self.req_cache.items()):
            req_type_str = req_type[0].upper(), req_type[1:].lower()
            monitor_entry.update({
                'sockCache%s%sNum' % req_type_str: [str(len(req_cache))],
                'sockCache%s%sTTL' % req_type_str: [ldap_float_str(req_cache._cache_ttl)],
                'sockCache%s%sMaxHitTime' % req_type_str: [
                    ldap_float_str(req_cache.max_cache_hit_time)
                ],
                'sockCache%s%sHitCount' % req_type_str: [str(req_cache.cache_hit_count)],
            })
            if self._req_counters[req_type]:
                monitor_entry['sockCache%s%sHitRate' % (
                    (req_type_str)
                )] = [
                    ldap_float_str(
                        100*float(req_cache.cache_hit_count) / \
                        float(self._req_counters[req_type])
                    )
                ]
            else:
                monitor_entry['sockCache%s%sHitRate' % (
                    (req_type_str)
                )] = [ldap_float_str(0.0)]
        return monitor_entry

    def disable_ldapi_conn(self):
        """
        Destroy local LDAPI connection and reset it to None.
        Should be invoked when catching a ldap.SERVER_DOWN exception.
        """
        try:
            self._ldapi_conn_lock.acquire()
            if self._ldapi_conn:
                self._ldapi_conn.unbind_s()
        finally:
            # Free it
            del self._ldapi_conn
            # Reset it
            self._ldapi_conn = None
            self._ldapi_conn_lock.release()

    def get_ldapi_conn(self):
        """
        Open a single local LDAPI connection and bind with SASL/EXTERNAL if
        needed
        """
        if isinstance(self._ldapi_conn, MyLDAPObject):
            self.logger.debug(
                'Existing LDAP connection to %s (%s)',
                repr(self._ldapi_conn._uri),
                repr(self._ldapi_conn),
            )
            return self._ldapi_conn
        try:
            self._ldapi_conn_lock.acquire()
            try:
                self._ldapi_conn = MyLDAPObject(
                    self.ldapi_uri,
                    trace_level=self.ldap_trace_level,
                    trace_file=self._logger_fileobj,
                    retry_max=self.ldap_retry_max,
                    retry_delay=self.ldap_retry_delay,
                    cache_time=self.ldap_cache_ttl,
                )
                self._ldapi_conn.sasl_external_bind_s(
                    authz_id=self.ldapi_authz_id or ''
                )
            except ldap.LDAPError, ldap_error:
                self._ldapi_conn = None
                self.logger.error(
                    'LDAPError during connecting to %s: %s',
                    repr(self.ldapi_uri),
                    str(ldap_error),
                )
                raise ldap_error
            else:
                self._ldapi_conn.authz_id = self._ldapi_conn.whoami_s()
                self.logger.info(
                    'Successfully bound to %s as %s (%s)',
                    repr(self.ldapi_uri),
                    repr(self._ldapi_conn.authz_id),
                    repr(self._ldapi_conn),
                )
        finally:
            self._ldapi_conn_lock.release()
        return self._ldapi_conn # get_ldapi_conn()

    def server_bind(self):
        """Override server_bind to set socket options."""
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.settimeout(self._socket_timeout)
        try:
            os.unlink(self.server_address)
        except OSError, os_error:
            if os.path.exists(self.server_address):
                raise os_error
        SocketServer.UnixStreamServer.server_bind(self)
        os.chmod(self.server_address, int(self._socket_permissions, 8))
        return  # server_bind()


class SlapdSockThreadingServer(SocketServer.ThreadingMixIn, SlapdSockServer):
    """
    This can be used to run as multi-threaded server
    """
    pass
