#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
slapd-sock listener demon which validates intercepted BIND requests
against local data (just a demo)
"""

import os, logging
import slapdsock

__version__ = slapdsock.__version__
__author__ = u'Michael Ströder <michael@stroeder.com>'

#-----------------------------------------------------------------------
# Configuration constants
#-----------------------------------------------------------------------

# UIDs and peer GIDS of peers which are granted access
# (list of int/strings)
ALLOWED_UIDS = [0, 'system_slapd', os.getuid()]
ALLOWED_GIDS = [0]

# String with octal representation of socket permissions
SOCKET_PERMISSIONS = '0666'

# Timeout in seconds for the server (Unix domain) socket
SOCKET_TIMEOUT = 20

# Logging formats
SYS_LOG_FORMAT = '%(levelname)s %(message)s'
CONSOLE_LOG_FORMAT = '%(asctime)s %(levelname)s %(message)s'

# Base number for floating average value of response delay
AVERAGE_COUNT = 100

# Default log level to use
LOG_LEVEL = logging.DEBUG

# Dictionary with bind-DN and passwords
USERPASSWORD_DICT = {
    u'uid=test1': 'foobar23',
    u'uid=test2': 'barfoo42',
    u'uid=äöüÄÖÜß': 'blrub',
}

#-----------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------

# from Python's standard lib
import sys

# local modules
from slapdsock.ldaphelper import RESULT_CODE
from slapdsock.loghelper import combined_logger
from slapdsock.handler import NoopHandler
from slapdsock.message import RESULTResponse

# run single-threaded
#from slapdsock.service import SlapdSockServer
# run multi-threaded
from slapdsock.service import SlapdSockThreadingServer as SlapdSockServer

#-----------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------


class FakeBindHandler(NoopHandler):

    """
    Handler class which validates bind-DN and password against
    local dictionary USERPASSWORD_DICT
    """

    users = USERPASSWORD_DICT

    def do_bind(self, request):
        bind_rdn = request.dn[:-len(request.suffix)-1]
        self.server.logger.debug(
            '%s bind_rdn=%s',
            self.log_prefix,
            bind_rdn,
        )
        try:
            user_password = self.users[bind_rdn]
        except KeyError:
            result_code = RESULT_CODE['invalidCredentials']
            info = 'You loose! (no such user)'
        else:
            if request.cred == user_password:
                result_code = RESULT_CODE['success']
                info = 'You won!'
            else:
                result_code = RESULT_CODE['invalidCredentials']
                info = 'You loose! (wrong password)'
        self.server.logger.info(
            '%s result_code=%d info="%s"',
            self.log_prefix,
            result_code,
            info,
        )
        return RESULTResponse(
            request.msgid,
            result_code,
            matched=request.dn,
            info=info,
        )
        # end of do_BIND()


#-----------------------------------------------------------------------
# Main
#-----------------------------------------------------------------------

script_name = sys.argv[0]

log_level = LOG_LEVEL
console_log_format = None
if __debug__ and os.environ.get('DEBUG', 'no') == 'yes':
    log_level = logging.DEBUG
    console_log_format = CONSOLE_LOG_FORMAT

my_logger = combined_logger(
    script_name,
    log_level,
    sys_log_format=SYS_LOG_FORMAT,
    console_log_format=console_log_format,
)

my_logger.info(
    'Starting %s %s (log level %d)',
    script_name,
    __version__,
    my_logger.level
)

try:
    socket_path = sys.argv[1]
except IndexError:
    my_logger.error('Not enough arguments => abort')
    sys.exit(1)

try:
    slapd_sock_listener = SlapdSockServer(
        socket_path,
        FakeBindHandler,
        my_logger,
        AVERAGE_COUNT,
        SOCKET_TIMEOUT, SOCKET_PERMISSIONS,
        ALLOWED_UIDS, ALLOWED_GIDS,
    )
    try:
        slapd_sock_listener.serve_forever()
    except KeyboardInterrupt:
        my_logger.warn('Received interrupt signal => shutdown')
finally:
    my_logger.debug('Remove socket path %s', repr(socket_path))
    try:
        os.remove(socket_path)
    except OSError:
        pass
